# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

import unittest
from tr55.water_quality import get_volume_of_runoff, get_pollutant_load
from tr55.tables import POLLUTION_LOADS


class TestWaterQuality(unittest.TestCase):

    def test_volume(self):
        """
        Test the water volume computation.
        """
        cell_res = 30  # meters
        cell_count = 100
        runoff = 0.4  # inches
        liters = get_volume_of_runoff(runoff, cell_count, cell_res)

        self.assertEquals(30480, liters,
                          "Did not calculate the correct runoff volume")

    def test_load(self):
        """
        Test the pollutant load computation.
        """
        nlcd = 24
        pollutant = 'tn'
        emc = POLLUTION_LOADS[nlcd][pollutant]
        runoff_liters = 1000

        expected = ((runoff_liters * emc) / 1000000) * 2.205
        load = get_pollutant_load('commercial', pollutant, runoff_liters)

        self.assertEquals(expected, load)

    def test_load_bad_nlcd(self):
        self.assertRaises(Exception, get_pollutant_load, 'asdf', 'tn', 1000)

    def test_load_bad_pollutant(self):
        self.assertRaises(Exception, get_pollutant_load, 'commercial',
                          'asdf', 1000)



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Various routines to do table lookups.
"""

from tr55.tables import SAMPLE_YEAR, BMPS, BUILT_TYPES, LAND_USE_VALUES, \
    PRE_COLUMBIAN_LAND_USES, POLLUTANTS, POLLUTION_LOADS


def lookup_ki(land_use):
    """
    Lookup the landuse coefficient.
    """
    if land_use not in LAND_USE_VALUES \
       or 'ki' not in LAND_USE_VALUES[land_use]:
        raise Exception('Unknown land use: %s' % land_use)
    else:
        return LAND_USE_VALUES[land_use]['ki']


def lookup_pet(day, land_use):
    """
    Lookup/compute evapotranspiration from the tables.
    """
    (precip, et_max) = SAMPLE_YEAR[day]
    ki = LAND_USE_VALUES[land_use]['ki']
    return (precip, et_max * ki)


def lookup_bmp_infiltration(soil_type, bmp):
    """
    Lookup the amount of infiltration causes by a particular BMP.
    """
    if bmp not in LAND_USE_VALUES \
       or 'infiltration' not in LAND_USE_VALUES[bmp]:
        raise Exception('%s not a BMP' % bmp)
    elif soil_type not in LAND_USE_VALUES[bmp]['infiltration']:
        raise Exception('BMP %s incompatible with soil %s' % (bmp, soil_type))
    else:
        return LAND_USE_VALUES[bmp]['infiltration'][soil_type]


def lookup_cn(soil_type, land_use):
    """
    Lookup the runoff curve number for a particular soil type and land use.
    """
    if land_use not in LAND_USE_VALUES:
        raise Exception('Unknown land use %s' % land_use)
    elif 'cn' not in LAND_USE_VALUES[land_use] \
         or soil_type not in LAND_USE_VALUES[land_use]['cn']:
        raise Exception('Unknown soil type %s' % soil_type)
    else:
        return LAND_USE_VALUES[land_use]['cn'][soil_type]


def is_bmp(land_use):
    """
    Test to see if the land use is a BMP.
    """
    return land_use in BMPS


def is_built_type(land_use):
    """
    Test to see if the land use is a "built type".
    """
    return land_use in BUILT_TYPES


def make_precolumbian(land_use):
    """
    Project the given land use to a Pre-Columbian one.
    """
    if land_use not in PRE_COLUMBIAN_LAND_USES:
        return 'mixed_forest'
    else:
        return land_use


def lookup_load(nlcd_class, pollutant):
    """
    Get the Event Mean Concentration of `pollutant` for land use
    class `nlcd_class`
    """
    if pollutant not in ['tn', 'tp', 'bod', 'tss']:
        raise Exception('Unknown pollutant type: %s' % pollutant)

    if nlcd_class not in POLLUTION_LOADS:
        raise Exception('Unknown NLCD class: %s' % nlcd_class)

    return POLLUTION_LOADS[nlcd_class][pollutant]


def lookup_nlcd(land_use):
    """
    Get the NLCD number for a particular human-readable land use.
    """
    if land_use not in LAND_USE_VALUES:
        raise Exception('Unknown land use type: %s' % land_use)

    if 'nlcd' not in LAND_USE_VALUES[land_use]:
        raise Exception('Land use type %s does not have an NLCD class defined',
                        land_use)

    return LAND_USE_VALUES[land_use]['nlcd']


def get_pollutants():
    """
    Return the list of pollutants.
    """
    return POLLUTANTS



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
TR-55 Model Implementation

A mapping between variable/parameter names found in the TR-55 document
and variables used in this program are as follows:
 * `precip` is referred to as P in the report
 * `runoff` is Q
 * `evaptrans` maps to ET, the evapotranspiration
 * `inf` is the amount of water that infiltrates into the soil (in inches)
 * `init_abs` is Ia, the initial abstraction, another form of infiltration
"""

import copy

from tr55.tablelookup import lookup_pet, lookup_cn, lookup_bmp_infiltration, \
    is_bmp, is_built_type, make_precolumbian, get_pollutants
from tr55.water_quality import get_volume_of_runoff, get_pollutant_load
from tr55.operations import dict_plus


def runoff_pitt(precip, land_use):
    """
    The Pitt Small Storm Hydrology method.  The output is a runoff
    value in inches.
    """
    c1 = +3.638858398e-2
    c2 = -1.243464039e-1
    c3 = +1.295682223e-1
    c4 = +9.375868043e-1
    c5 = -2.235170859e-2
    c6 = +0.170228067e0
    c7 = -3.971810782e-1
    c8 = +3.887275538e-1
    c9 = -2.289321859e-2
    p4 = pow(precip, 4)
    p3 = pow(precip, 3)
    p2 = pow(precip, 2)

    impervious = (c1 * p3) + (c2 * p2) + (c3 * precip) + c4
    urb_grass = (c5 * p4) + (c6 * p3) + (c7 * p2) + (c8 * precip) + c9

    runoff_vals = {
        'water':           impervious,
        'li_residential':  0.20 * impervious + 0.80 * urb_grass,
        'cluster_housing': 0.20 * impervious + 0.80 * urb_grass,
        'hi_residential':  0.65 * impervious + 0.35 * urb_grass,
        'commercial':      impervious,
        'industrial':      impervious,
        'transportation':  impervious,
        'urban_grass':     urb_grass
    }

    if land_use not in runoff_vals:
        raise Exception('Land use %s not a built-type.' % land_use)
    else:
        return min(runoff_vals[land_use], precip)


def nrcs_cutoff(precip, curve_number):
    """
    A function to find the cutoff between preciptation/curve number
    pairs that have zero runoff by definition, and those that do not.
    """
    if precip <= -1 * (2 * (curve_number - 100.0) / curve_number):
        return True
    else:
        return False


def runoff_nrcs(precip, evaptrans, soil_type, land_use):
    """
    The runoff equation from the TR-55 document.  The output is a
    runoff value in inches.
    """
    curve_number = lookup_cn(soil_type, land_use)
    if nrcs_cutoff(precip, curve_number):
        return 0.0
    potential_retention = (1000.0 / curve_number) - 10
    initial_abs = 0.2 * potential_retention
    precip_minus_initial_abs = precip - initial_abs
    numerator = pow(precip_minus_initial_abs, 2)
    denominator = (precip_minus_initial_abs + potential_retention)
    runoff = numerator / denominator
    return min(runoff, precip - evaptrans)


def simulate_cell_day(parameters, cell, cell_count):
    """
    Simulate a bunch of cells of the same type during a one-day event.

    The first argument, `parameters`, contains the precipitation and
    evapotranspiration as a tuple.

    `cell` is a string which contains a soil type and land use
    separated by a colon.

    `cell_count` is the number of cells to simulate.

    The return value is a dictionary of runoff, evapotranspiration, and
    infiltration as volumes of water.
    """
    precip, evaptrans = parameters
    soil_type, land_use = cell.lower().split(':')

    # If there is no precipitation, then there is no runoff or
    # infiltration.  There is evapotranspiration, however (is
    # understood that over a period of time, this can lead to the sum
    # of the three values exceeding the total precipitation).
    if precip == 0.0:
        return {
            'runoff-vol': 0.0,
            'et-vol': cell_count * evaptrans,
            'inf-vol': 0.0
        }

    # Deal with the Best Management Practices (BMPs).  For most BMPs,
    # the infiltration is read from the table and the runoff is what
    # is left over after infiltration and evapotranspiration.  Rain
    # gardens are treated differently.
    if is_bmp(land_use) and land_use != 'rain_garden':
        inf = lookup_bmp_infiltration(soil_type, land_use)  # infiltration
        runoff = precip - (evaptrans + inf)  # runoff
        return {
            'runoff-vol': cell_count * runoff,
            'et-vol': cell_count * evaptrans,
            'inf-vol': cell_count * inf
        }
    elif land_use == 'rain_garden':
        # Here, return a mixture of 20% ideal rain garden and 80% high
        # intensity residential.
        inf = lookup_bmp_infiltration(soil_type, land_use)
        runoff = precip - (evaptrans + inf)
        hi_res_cell = soil_type + ':hi_residential'
        hi_res = simulate_cell_day((precip, evaptrans), hi_res_cell, 1)
        return {
            'runoff-vol': cell_count * (0.2 * runoff + 0.8 * hi_res[0]),
            'et-vol': cell_count * (0.2 * evaptrans + 0.8 * hi_res[1]),
            'inf-vol': cell_count * (0.2 * inf + 0.8 * hi_res[2])
        }

    # When the land use is a built-type and the level of precipitation
    # is two inches or less, use the Pitt Small Storm Hydrology Model.
    # When the land use is a built-type but the level of precipitation
    # is higher, the runoff is the larger of that predicted by the
    # Pitt model and NRCS model.  Otherwise, return the NRCS amount.
    if is_built_type(land_use) and precip <= 2.0:
        runoff = runoff_pitt(precip, land_use)
    elif is_built_type(land_use):
        pitt_runoff = runoff_pitt(2.0, land_use)
        nrcs_runoff = runoff_nrcs(precip, evaptrans, soil_type, land_use)
        runoff = max(pitt_runoff, nrcs_runoff)
    else:
        runoff = runoff_nrcs(precip, evaptrans, soil_type, land_use)
    inf = precip - (evaptrans + runoff)

    return {
        'runoff-vol': cell_count * runoff,
        'et-vol': cell_count * evaptrans,
        'inf-vol': cell_count * max(inf, 0.0),
    }


def simulate_cell_year(cell, cell_count, precolumbian):
    """
    Simulate a cell-type for an entire year using sample precipitation and
    evapotranspiration data.

    The `cell` parameter is a string with the soil type and land use
    separated by a colon.

    The `cell_count` parameter is the number of cells of this type.

    If the `precolumbian` parameter is true, then the cell is
    simulated under Pre-Columbian circumstances (anything other than
    water, woody wetland, and herbaceous wetland becomes mixed
    forest).
    """
    (soil_type, land_use) = cell.split(':')

    if precolumbian:
        land_use = make_precolumbian(land_use)
        cell = soil_type + ':' + land_use

    retval = {}
    for day in range(365):
        pet = lookup_pet(day, land_use)
        retval = dict_plus(retval, simulate_cell_day(pet, cell, cell_count))
    return retval


def create_unmodified_census(census):
    """
    This creates a cell census, ignoring any modifications.  The
    output is suitable for use with `simulate_water_quality`.
    """
    unmod = copy.deepcopy(census)
    unmod.pop('modifications', None)
    return unmod


def create_modified_census(census):
    """
    This creates a cell census, with modifications, that is suitable
    for use with `simulate_water_quality`.

    For every type of cell that undergoes modification, the
    modifications are indicated with a sub-distribution under that
    cell type.
    """
    mod = copy.deepcopy(census)
    mod.pop('modifications', None)

    if 'modifications' in census:
        for change in census['modifications']:
            for (cell, subcensus) in change['distribution'].items():
                parent = mod['distribution'][cell]
                n = parent['cell_count']
                m = subcensus['cell_count']

                if 'bmp' in change:
                    changed_cell = cell.split(':')[0] + ':' + change['bmp']
                elif 'reclassification' in change:
                    changed_cell = change['reclassification']
                else:
                    raise Exception('Unknown modification type.')

                parent['distribution'] = {
                    cell: {"cell_count": n - m},
                    changed_cell: {"cell_count": m}
                }

    return mod


def simulate_water_quality(tree, cell_res, fn,
                           parent_cell=None, current_cell=None):
    """
    Perform a water quality simulation by doing simulations on each
    type of cells (leaves), then adding them together going upward
    (summing the values of a node's subtrees and storing them at that
    node).

    `parent_cell` is the cell type (a string with a soil type and land
    use separated by a colon) of the parent of the present node in
    tree.

    `current_cell` is the cell type for the present node.

    `cell_res` is the size of each cell (used for turning inches of
    water into volumes of water).

    `tree` is the (sub)tree of cell distributions that is currently
    under consideration.

    `fn` is a function that takes a cell type and a number of cells
    and returns a dictionary containing runoff, et, and inf as
    volumes.  This typically just calls `simulate_cell_year`, but can
    be set to something else if e.g. a simulation over a different
    time-scale is desired.
    """
    # Internal node.
    if 'cell_count' in tree and 'distribution' in tree:
        # simulate subtrees
        tally = {}
        for cell, subtree in tree['distribution'].items():
            simulate_water_quality(subtree, cell_res, fn, current_cell, cell)
            subtree_ex_dist = subtree.copy()
            subtree_ex_dist.pop('distribution', None)
            tally = dict_plus(tally, subtree_ex_dist)

        # update this node
        tree.update(tally)

    # Leaf node.
    elif 'cell_count' in tree and 'distribution' not in tree:
        # runoff, et, inf
        n = tree['cell_count']
        result = fn(current_cell, n)
        tree.update(result)

        # water quality
        if n != 0:
            runoff = result['runoff-vol'] / n
            liters = get_volume_of_runoff(runoff, n, cell_res)
            land_use = current_cell.split(':')[1]
            if is_bmp(land_use) or land_use == 'no_till' or \
               land_use == 'cluster_housing':
                land_use = parent_cell.split(':')[1]
            for pol in get_pollutants():
                tree[pol] = get_pollutant_load(land_use, pol, liters)


def postpass(tree):
    """
    Remove volume units and replace them with inches.
    """
    if 'cell_count' in tree:
        n = tree['cell_count']
        tree['runoff'] = tree['runoff-vol'] / n
        tree['et'] = tree['et-vol'] / n
        tree['inf'] = tree['inf-vol'] / n
        tree.pop('runoff-vol', None)
        tree.pop('et-vol', None)
        tree.pop('inf-vol', None)
    if 'distribution' in tree:
        for subtree in tree['distribution'].values():
            postpass(subtree)


def simulate_modifications(census, cell_res=10, precolumbian=False, fn=None):
    """
    Simulate an entire year, including effects of modifications.

    `census` contains a distribution of cell-types in the area of interest.

    `cell_res` is as described in `simulate_water_quality`.

    `precolumbian` is as described in `simulate_cell_year`.

    `fn` is as described in `simulate_water_quality`.  If this is
    supplied, then `cell_res` and `precolumbian` are ignored.
    """
    if not fn:
        def fn(cell, cell_count):
            return simulate_cell_year(cell, cell_count, precolumbian)

    mod = create_modified_census(census)
    simulate_water_quality(mod, cell_res, fn)
    postpass(mod)

    unmod = create_unmodified_census(census)
    simulate_water_quality(unmod, cell_res, fn)
    postpass(unmod)

    return {
        'unmodified': unmod,
        'modified': mod
    }



"""
Generates a table with the model output for different input values.
This table is used in the mini-app.
"""

from itertools import product
import csv
import sys

from tr55.model import simulate_modifications, simulate_cell_day
from tr55.tablelookup import lookup_ki

if len(sys.argv) != 2:
    print 'Usage: python -m tr55.makeMiniAppTable csv_file_name'
    sys.exit()
else:
    csv_file_name = sys.argv[1]

with open(csv_file_name, 'wb') as csv_file:
    writer = csv.writer(csv_file)
    writer.writerow(('P', 'land', 'soil', 'ET', 'I', 'R'))

    # values of inputs to the model
    precips = [0.5, 1.0, 2.0, 3.2, 8.0]

    # The land uses in the original mini-app were low intensity
    # residential, high intensity residential, commercial, grassland, forest,
    # turf grass, pasture, row crops. The closest matching to the values in
    # this implementation of TR-55 are the following:
    land_uses = [
        'li_residential',
        'hi_residential',
        'commercial',
        'grassland',
        'mixed_forest',
        'urban_grass',
        'pasture',
        'row_crop',
        'chaparral',
        'tall_grass_prairie',
        'short_grass_prairie',
        'desert'
    ]

    soil_types = ['a', 'b', 'c', 'd']

    et_max = 0.207
    pre_columbian = False

    # For each input value, compute the model outputs for a
    # single day and tile.
    for precip, land_use, soil_type in product(precips,
                                               land_uses,
                                               soil_types):
        cells = {
            'cell_count': 1,
            'distribution': {
                '%s:%s' % (soil_type, land_use): {'cell_count': 1}
            }
        }

        def fn(cell, cell_count):
            (soil_type, land_use) = cell.lower().split(':')
            et = et_max * lookup_ki(land_use)
            return simulate_cell_day((precip, et), cell, cell_count)

        model_out = (simulate_modifications(cells, fn=fn)
                     ['unmodified']['distribution'].values()[0])
        writer.writerow((precip, land_use,
                         soil_type, model_out['et'],
                         model_out['inf'], model_out['runoff']))



"""
Compares the tables generated by makeMiniAppTable and the
 corresponding table from the original mini app.
"""

import csv
import sys

from tr55.tablelookup import lookup_nlcd


# Generate keys from rows that will be used to match rows in the
# old and new table.
# If transform == True, convert values of row so that
# land and soil_types are in the format of the data table
# from the old mini-app
def get_key(row, transform=False):
    P = row['P']
    land = row['land']
    soil = row['soil']
    if transform:
        land = str(lookup_nlcd(row['land']))
        new_soil_types = ['a', 'b', 'c', 'd']
        old_soil_types = ['0', '1', '2', '3']
        soil_types_map = dict(zip(new_soil_types, old_soil_types))
        soil = soil_types_map[row['soil']]

    return (P, land, soil)


def isClose(val1, val2):
    val1 = round(float(val1), 1)
    val2 = round(float(val2), 1)
    return val1 == val2


def rowsMatch(old_row, new_row):
    return isClose(old_row['ET'], new_row['ET']) and \
        isClose(old_row['I'], new_row['I']) and \
        isClose(old_row['R'], new_row['R'])


def test_get_key():
    old_row = {'land': '71', 'I': '1.6', 'soil': '2',
               'P': '2.0', 'R': '0.2', 'ET': '0.1'}
    new_row = {'land': 'grassland', 'I': '1.6100773216841155',
               'soil': 'c', 'P': '2.0', 'R': '0.2657226783158845',
               'ET': '0.12419999999999999'}
    assert get_key(old_row) == get_key(new_row, transform=True) \
                            == ('2.0', '71', '2')


test_get_key()

if len(sys.argv) != 3:
    print ('Usage: python -m tr55.compareMiniAppTable ' +
           'old_csv_file_name new_csv_file_name')
    sys.exit()
else:
    old_csv_fn = sys.argv[1]
    new_csv_fn = sys.argv[2]

with open(old_csv_fn, 'rb') as old_csv_file:
    with open(new_csv_fn, 'rb') as new_csv_file:
        old_reader = csv.DictReader(old_csv_file)
        new_reader = csv.DictReader(new_csv_file)

        old_row_dict = dict([(get_key(old_row), old_row)
                             for old_row in old_reader])
        new_row_dict = dict([(get_key(new_row, transform=True), new_row)
                             for new_row in new_reader])

        num_in_match_rows = 0
        num_in_out_match_rows = 0
        for key, new_row in new_row_dict.iteritems():
            if key in old_row_dict:
                old_row = old_row_dict[key]
                if rowsMatch(old_row, new_row):
                    num_in_out_match_rows += 1
                else:
                    print 'output does not match'
                    print 'old row: ', old_row
                    print 'new row: ', new_row
                    print
                num_in_match_rows += 1

        print '# rows in new table: ', len(new_row_dict)
        print '# rows where input matches: ', num_in_match_rows
        print '# rows where input and output match: ', num_in_out_match_rows



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Table Lookup test set.
"""

import unittest

from tr55.tablelookup import lookup_pet, lookup_bmp_infiltration, lookup_cn


class TestTablelookups(unittest.TestCase):
    """
    Table Lookup test set
    """
    def test_lookup_pet(self):
        """
        Do some spot-checks on the SampleYear data.
        """
        self.assertEqual(lookup_pet(0, 'water')[0], 0.0)
        self.assertEqual(lookup_pet(52, 'hi_residential')[1], 0.0)
        self.assertEqual(lookup_pet(104, 'green_roof')[1], 0.0)
        self.assertEqual(lookup_pet(156, 'cluster_housing')[0], 0.23)
        self.assertEqual(lookup_pet(208, 'grassland')[0], 0.2)
        self.assertEqual(lookup_pet(260, 'woody_wetland')[1], 0.207)
        self.assertEqual(lookup_pet(352, 'hay')[1], 0.207 * 0.6)

    def test_lookup_bmp_infiltration(self):
        """
        Do some spot-checks on the data from Table B.
        """
        self.assertEqual(lookup_bmp_infiltration('d', 'green_roof'), 1.6)
        self.assertEqual(lookup_bmp_infiltration('c', 'porous_paving'), 1.73)
        self.assertEqual(lookup_bmp_infiltration('b', 'rain_garden'), 0.6)
        self.assertEqual(lookup_bmp_infiltration('a', 'infiltration_trench'), 2.4)  # noqa

    def test_lookup_cn(self):
        """
        Do some spot-checks on the data from Table C.
        """
        self.assertEqual(lookup_cn('a', 'water'), 100)
        self.assertEqual(lookup_cn('b', 'li_residential'), 68)
        self.assertEqual(lookup_cn('c', 'hi_residential'), 90)
        self.assertEqual(lookup_cn('d', 'commercial'), 95)
        self.assertEqual(lookup_cn('a', 'rock'), 77)
        self.assertEqual(lookup_cn('b', 'deciduous_forest'), 55)
        self.assertEqual(lookup_cn('c', 'evergreen_forest'), 70)
        self.assertEqual(lookup_cn('d', 'mixed_forest'), 77)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Model test set
"""

import unittest
from math import sqrt

from tr55.model import runoff_nrcs, \
    simulate_cell_day, simulate_cell_year, simulate_water_quality, \
    create_unmodified_census, create_modified_census
from tr55.tablelookup import lookup_ki

# These data are taken directly from Table 2-1 of the revised (1986)
# TR-55 report.  The data in the PS array are various precipitation
# levels, and each respective CNx array is the calculated runoff for
# that particular curve number with the given level of precipitation
# corresponding to that in PS.
PS = [1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0]  # noqa
CN55 = [0.000, 0.000, 0.000, 0.000, 0.000, 0.020, 0.080, 0.190, 0.350, 0.530, 0.740, 0.980, 1.520, 2.120, 2.780, 3.490, 4.230, 5.000, 5.790, 6.610, 7.440, 8.290]  # noqa
CN70 = [0.000, 0.030, 0.060, 0.110, 0.170, 0.240, 0.460, 0.710, 1.010, 1.330, 1.670, 2.040, 2.810, 3.620, 4.460, 5.330, 6.220, 7.130, 8.050, 8.980, 9.910, 10.85]  # noqa
CN80 = [0.080, 0.150, 0.240, 0.340, 0.440, 0.560, 0.890, 1.250, 1.640, 2.040, 2.460, 2.890, 3.780, 4.690, 5.630, 6.570, 7.520, 8.480, 9.450, 10.42, 11.39, 12.37]  # noqa
CN90 = [0.320, 0.460, 0.610, 0.760, 0.930, 1.090, 1.530, 1.980, 2.450, 2.920, 3.400, 3.880, 4.850, 5.820, 6.810, 7.790, 8.780, 9.770, 10.76, 11.76, 12.75, 13.74]  # noqa

# INPUT and OUTPUT are data that were emailed to Azavea in a
# spreadsheet for testing the TR-55 model implementation.
INPUT = [
    (0.5, 'a:water'),
    (1, 'a:water'),
    (2, 'a:water'),
    (3.2, 'a:water'),
    (8, 'a:water'),
    (0.5, 'a:rock'),
    (1, 'a:rock'),
    (2, 'a:rock'),
    (3.2, 'a:rock'),
    (8, 'a:rock'),
    (0.5, 'a:urban_grass'),
    (1, 'a:urban_grass'),
    (2, 'a:urban_grass'),
    (3.2, 'a:urban_grass'),
    (8, 'a:urban_grass'),
    (0.5, 'a:li_residential'),
    (1, 'a:li_residential'),
    (2, 'a:li_residential'),
    (3.2, 'a:li_residential'),
    (8, 'a:li_residential'),
    (0.5, 'a:hi_residential'),
    (1, 'a:hi_residential'),
    (2, 'a:hi_residential'),
    (3.2, 'a:hi_residential'),
    (8, 'a:hi_residential'),
    (0.5, 'a:commercial'),
    (1, 'a:commercial'),
    (2, 'a:commercial'),
    (3.2, 'a:commercial'),
    (8, 'a:commercial'),
    (0.5, 'a:deciduous_forest'),
    (0.5, 'a:evergreen_forest'),
    (0.5, 'a:mixed_forest'),
    (1, 'a:deciduous_forest'),
    (1, 'a:evergreen_forest'),
    (1, 'a:mixed_forest'),
    (2, 'a:deciduous_forest'),
    (2, 'a:evergreen_forest'),
    (2, 'a:mixed_forest'),
    (3.2, 'a:deciduous_forest'),
    (3.2, 'a:evergreen_forest'),
    (3.2, 'a:mixed_forest'),
    (8, 'a:deciduous_forest'),
    (8, 'a:evergreen_forest'),
    (8, 'a:mixed_forest'),
    (0.5, 'a:grassland'),
    (1, 'a:grassland'),
    (2, 'a:grassland'),
    (3.2, 'a:grassland'),
    (8, 'a:grassland'),
    (0.5, 'a:pasture'),
    (1, 'a:pasture'),
    (2, 'a:pasture'),
    (3.2, 'a:pasture'),
    (8, 'a:pasture'),
    (0.5, 'a:row_crop'),
    (1, 'a:row_crop'),
    (2, 'a:row_crop'),
    (3.2, 'a:row_crop'),
    (8, 'a:row_crop'),
    (0.5, 'a:woody_wetland'),
    (0.5, 'a:herbaceous_wetland'),
    (1, 'a:woody_wetland'),
    (1, 'a:herbaceous_wetland'),
    (2, 'a:woody_wetland'),
    (2, 'a:herbaceous_wetland'),
    (3.2, 'a:woody_wetland'),
    (3.2, 'a:herbaceous_wetland'),
    (8, 'a:woody_wetland'),
    (8, 'a:herbaceous_wetland'),
    (0.5, 'b:water'),
    (1, 'b:water'),
    (2, 'b:water'),
    (3.2, 'b:water'),
    (8, 'b:water'),
    (0.5, 'b:rock'),
    (1, 'b:rock'),
    (2, 'b:rock'),
    (3.2, 'b:rock'),
    (8, 'b:rock'),
    (0.5, 'b:urban_grass'),
    (1, 'b:urban_grass'),
    (2, 'b:urban_grass'),
    (3.2, 'b:urban_grass'),
    (8, 'b:urban_grass'),
    (0.5, 'b:li_residential'),
    (1, 'b:li_residential'),
    (2, 'b:li_residential'),
    (3.2, 'b:li_residential'),
    (8, 'b:li_residential'),
    (0.5, 'b:hi_residential'),
    (1, 'b:hi_residential'),
    (2, 'b:hi_residential'),
    (3.2, 'b:hi_residential'),
    (8, 'b:hi_residential'),
    (0.5, 'b:commercial'),
    (1, 'b:commercial'),
    (2, 'b:commercial'),
    (3.2, 'b:commercial'),
    (8, 'b:commercial'),
    (0.5, 'b:deciduous_forest'),
    (0.5, 'b:evergreen_forest'),
    (0.5, 'b:mixed_forest'),
    (1, 'b:deciduous_forest'),
    (1, 'b:evergreen_forest'),
    (1, 'b:mixed_forest'),
    (2, 'b:deciduous_forest'),
    (2, 'b:evergreen_forest'),
    (2, 'b:mixed_forest'),
    (3.2, 'b:deciduous_forest'),
    (3.2, 'b:evergreen_forest'),
    (3.2, 'b:mixed_forest'),
    (8, 'b:deciduous_forest'),
    (8, 'b:evergreen_forest'),
    (8, 'b:mixed_forest'),
    (0.5, 'b:grassland'),
    (1, 'b:grassland'),
    (2, 'b:grassland'),
    (3.2, 'b:grassland'),
    (8, 'b:grassland'),
    (0.5, 'b:pasture'),
    (1, 'b:pasture'),
    (2, 'b:pasture'),
    (3.2, 'b:pasture'),
    (8, 'b:pasture'),
    (0.5, 'b:row_crop'),
    (1, 'b:row_crop'),
    (2, 'b:row_crop'),
    (3.2, 'b:row_crop'),
    (8, 'b:row_crop'),
    (0.5, 'b:woody_wetland'),
    (0.5, 'b:herbaceous_wetland'),
    (1, 'b:woody_wetland'),
    (1, 'b:herbaceous_wetland'),
    (2, 'b:woody_wetland'),
    (2, 'b:herbaceous_wetland'),
    (3.2, 'b:woody_wetland'),
    (3.2, 'b:herbaceous_wetland'),
    (8, 'b:woody_wetland'),
    (8, 'b:herbaceous_wetland'),
    (0.5, 'c:water'),
    (1, 'c:water'),
    (2, 'c:water'),
    (3.2, 'c:water'),
    (8, 'c:water'),
    (0.5, 'c:rock'),
    (1, 'c:rock'),
    (2, 'c:rock'),
    (3.2, 'c:rock'),
    (8, 'c:rock'),
    (0.5, 'c:urban_grass'),
    (1, 'c:urban_grass'),
    (2, 'c:urban_grass'),
    (3.2, 'c:urban_grass'),
    (8, 'c:urban_grass'),
    (0.5, 'c:li_residential'),
    (1, 'c:li_residential'),
    (2, 'c:li_residential'),
    (3.2, 'c:li_residential'),
    (8, 'c:li_residential'),
    (0.5, 'c:hi_residential'),
    (1, 'c:hi_residential'),
    (2, 'c:hi_residential'),
    (3.2, 'c:hi_residential'),
    (8, 'c:hi_residential'),
    (0.5, 'c:commercial'),
    (1, 'c:commercial'),
    (2, 'c:commercial'),
    (3.2, 'c:commercial'),
    (8, 'c:commercial'),
    (0.5, 'c:deciduous_forest'),
    (0.5, 'c:evergreen_forest'),
    (0.5, 'c:mixed_forest'),
    (1, 'c:deciduous_forest'),
    (1, 'c:evergreen_forest'),
    (1, 'c:mixed_forest'),
    (2, 'c:deciduous_forest'),
    (2, 'c:evergreen_forest'),
    (2, 'c:mixed_forest'),
    (3.2, 'c:deciduous_forest'),
    (3.2, 'c:evergreen_forest'),
    (3.2, 'c:mixed_forest'),
    (8, 'c:deciduous_forest'),
    (8, 'c:evergreen_forest'),
    (8, 'c:mixed_forest'),
    (0.5, 'c:grassland'),
    (1, 'c:grassland'),
    (2, 'c:grassland'),
    (3.2, 'c:grassland'),
    (8, 'c:grassland'),
    (0.5, 'c:pasture'),
    (1, 'c:pasture'),
    (2, 'c:pasture'),
    (3.2, 'c:pasture'),
    (8, 'c:pasture'),
    (0.5, 'c:row_crop'),
    (1, 'c:row_crop'),
    (2, 'c:row_crop'),
    (3.2, 'c:row_crop'),
    (8, 'c:row_crop'),
    (0.5, 'c:woody_wetland'),
    (0.5, 'c:herbaceous_wetland'),
    (1, 'c:woody_wetland'),
    (1, 'c:herbaceous_wetland'),
    (2, 'c:woody_wetland'),
    (2, 'c:herbaceous_wetland'),
    (3.2, 'c:woody_wetland'),
    (3.2, 'c:herbaceous_wetland'),
    (8, 'c:woody_wetland'),
    (8, 'c:herbaceous_wetland'),
    (0.5, 'd:water'),
    (1, 'd:water'),
    (2, 'd:water'),
    (3.2, 'd:water'),
    (8, 'd:water'),
    (0.5, 'd:rock'),
    (1, 'd:rock'),
    (2, 'd:rock'),
    (3.2, 'd:rock'),
    (8, 'd:rock'),
    (0.5, 'd:urban_grass'),
    (1, 'd:urban_grass'),
    (2, 'd:urban_grass'),
    (3.2, 'd:urban_grass'),
    (8, 'd:urban_grass'),
    (0.5, 'd:li_residential'),
    (1, 'd:li_residential'),
    (2, 'd:li_residential'),
    (3.2, 'd:li_residential'),
    (8, 'd:li_residential'),
    (0.5, 'd:hi_residential'),
    (1, 'd:hi_residential'),
    (2, 'd:hi_residential'),
    (3.2, 'd:hi_residential'),
    (8, 'd:hi_residential'),
    (0.5, 'd:commercial'),
    (1, 'd:commercial'),
    (2, 'd:commercial'),
    (3.2, 'd:commercial'),
    (8, 'd:commercial'),
    (0.5, 'd:deciduous_forest'),
    (0.5, 'd:evergreen_forest'),
    (0.5, 'd:mixed_forest'),
    (1, 'd:deciduous_forest'),
    (1, 'd:evergreen_forest'),
    (1, 'd:mixed_forest'),
    (2, 'd:deciduous_forest'),
    (2, 'd:evergreen_forest'),
    (2, 'd:mixed_forest'),
    (3.2, 'd:deciduous_forest'),
    (3.2, 'd:evergreen_forest'),
    (3.2, 'd:mixed_forest'),
    (8, 'd:deciduous_forest'),
    (8, 'd:evergreen_forest'),
    (8, 'd:mixed_forest'),
    (0.5, 'd:grassland'),
    (1, 'd:grassland'),
    (2, 'd:grassland'),
    (3.2, 'd:grassland'),
    (8, 'd:grassland'),
    (0.5, 'd:pasture'),
    (1, 'd:pasture'),
    (2, 'd:pasture'),
    (3.2, 'd:pasture'),
    (8, 'd:pasture'),
    (0.5, 'd:row_crop'),
    (1, 'd:row_crop'),
    (2, 'd:row_crop'),
    (3.2, 'd:row_crop'),
    (8, 'd:row_crop'),
    (0.5, 'd:woody_wetland'),
    (0.5, 'd:herbaceous_wetland'),
    (1, 'd:woody_wetland'),
    (1, 'd:herbaceous_wetland'),
    (2, 'd:woody_wetland'),
    (2, 'd:herbaceous_wetland'),
    (3.2, 'd:woody_wetland'),
    (3.2, 'd:herbaceous_wetland'),
    (8, 'd:woody_wetland'),
    (8, 'd:herbaceous_wetland')
]

OUTPUT = [
    (0.5, 0, 0),
    (1, 0, 0),
    (2, 0, 0),
    (3.2, 0, 0),
    (8, 0, 0),
    (0, 0, 0.5),
    (0, 0, 1),
    (0.4, 0, 1.6),
    (1.2, 0, 2),
    (5.3, 0, 2.7),
    (0, 0.1, 0.3),
    (0.1, 0.1, 0.7),
    (0.2, 0.1, 1.7),
    (0.7, 0.1, 2.3),
    (4.2, 0.1, 3.6),
    (0.1, 0.1, 0.3),
    (0.3, 0.1, 0.6),
    (0, 0.1, 1.9),
    (0.2, 0.1, 3),
    (2.4, 0.1, 5.6),
    (0.3, 0, 0.1),
    (0.7, 0, 0.3),
    (0.4, 0, 1.5),
    (1.2, 0, 2),
    (5.3, 0, 2.7),
    (0.5, 0, 0),
    (1, 0, 0),
    (1, 0, 1),
    (2.1, 0, 1.1),
    (6.7, 0, 1.3),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0, 0.1, 0.9),
    (0, 0.1, 0.9),
    (0, 0.1, 1.9),
    (0, 0.1, 1.9),
    (0, 0.1, 1.9),
    (0, 0.1, 3.1),
    (0, 0.1, 3.1),
    (0, 0.1, 3.1),
    (0.4, 0.1, 7.4),
    (0.4, 0.1, 7.4),
    (0.4, 0.1, 7.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0, 0.1, 1.9),
    (0, 0.1, 3.1),
    (0.4, 0.1, 7.5),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0, 0.1, 1.9),
    (0, 0.1, 3.1),
    (1.2, 0.1, 6.7),
    (0, 0.2, 0.3),
    (0, 0.2, 0.8),
    (0.2, 0.2, 1.6),
    (0.7, 0.2, 2.3),
    (4.1, 0.2, 3.7),
    (0.3, 0.2, 0),
    (0.3, 0.2, 0),
    (0.8, 0.2, 0),
    (0.8, 0.2, 0),
    (1.8, 0.2, 0),
    (1.8, 0.2, 0),
    (3, 0.2, 0),
    (3, 0.2, 0),
    (7.8, 0.2, 0),
    (7.8, 0.2, 0),
    (0.5, 0, 0),
    (1, 0, 0),
    (2, 0, 0),
    (3.2, 0, 0),
    (8, 0, 0),
    (0, 0, 0.5),
    (0.2, 0, 0.8),
    (0.8, 0, 1.2),
    (1.8, 0, 1.4),
    (6.3, 0, 1.7),
    (0, 0.1, 0.3),
    (0.1, 0.1, 0.7),
    (0.5, 0.1, 1.3),
    (1.3, 0.1, 1.7),
    (5.5, 0.1, 2.3),
    (0.1, 0.1, 0.3),
    (0.3, 0.1, 0.6),
    (0.2, 0.1, 1.7),
    (0.7, 0.1, 2.4),
    (4.2, 0.1, 3.7),
    (0.3, 0, 0.1),
    (0.7, 0, 0.3),
    (0.8, 0, 1.2),
    (1.8, 0, 1.4),
    (6.2, 0, 1.7),
    (0.5, 0, 0),
    (1, 0, 0),
    (1.2, 0, 0.8),
    (2.4, 0, 0.8),
    (7, 0, 0.9),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0, 0.1, 0.9),
    (0, 0.1, 0.9),
    (0, 0.1, 1.8),
    (0, 0.1, 1.8),
    (0, 0.1, 1.8),
    (0.3, 0.1, 2.8),
    (0.3, 0.1, 2.8),
    (0.3, 0.1, 2.8),
    (2.8, 0.1, 5.1),
    (2.8, 0.1, 5.1),
    (2.8, 0.1, 5.1),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0, 0.1, 1.9),
    (0.3, 0.1, 2.8),
    (2.8, 0.1, 5.1),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0.1, 0.1, 1.8),
    (0.4, 0.1, 2.6),
    (3.4, 0.1, 4.4),
    (0, 0.2, 0.3),
    (0.1, 0.2, 0.8),
    (0.5, 0.2, 1.3),
    (1.3, 0.2, 1.7),
    (5.4, 0.2, 2.4),
    (0.3, 0.2, 0),
    (0.3, 0.2, 0),
    (0.8, 0.2, 0),
    (0.8, 0.2, 0),
    (1.8, 0.2, 0),
    (1.8, 0.2, 0),
    (3, 0.2, 0),
    (3, 0.2, 0),
    (7.8, 0.2, 0),
    (7.8, 0.2, 0),
    (0.5, 0, 0),
    (1, 0, 0),
    (2, 0, 0),
    (3.2, 0, 0),
    (8, 0, 0),
    (0.1, 0, 0.4),
    (0.4, 0, 0.6),
    (1.2, 0, 0.8),
    (2.3, 0, 0.9),
    (6.9, 0, 1.1),
    (0, 0.1, 0.3),
    (0.1, 0.1, 0.7),
    (0.8, 0.1, 1),
    (1.8, 0.1, 1.2),
    (6.3, 0.1, 1.5),
    (0.1, 0.1, 0.3),
    (0.3, 0.1, 0.6),
    (0.5, 0.1, 1.4),
    (1.3, 0.1, 1.8),
    (5.5, 0.1, 2.4),
    (0.3, 0, 0.1),
    (0.7, 0, 0.3),
    (1.1, 0, 0.9),
    (2.2, 0, 1),
    (6.8, 0, 1.2),
    (0.5, 0, 0),
    (1, 0, 0),
    (1.4, 0, 0.6),
    (2.5, 0, 0.6),
    (7.3, 0, 0.7),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.8),
    (0, 0.1, 0.8),
    (0, 0.1, 0.8),
    (0.2, 0.1, 1.6),
    (0.2, 0.1, 1.6),
    (0.2, 0.1, 1.6),
    (0.8, 0.1, 2.2),
    (0.8, 0.1, 2.2),
    (0.8, 0.1, 2.2),
    (4.5, 0.1, 3.4),
    (4.5, 0.1, 3.4),
    (4.5, 0.1, 3.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0.2, 0.1, 1.6),
    (0.8, 0.1, 2.2),
    (4.5, 0.1, 3.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.9),
    (0.3, 0.1, 1.5),
    (1, 0.1, 2),
    (4.9, 0.1, 2.9),
    (0, 0.2, 0.3),
    (0.2, 0.2, 0.6),
    (0.8, 0.2, 1),
    (1.8, 0.2, 1.3),
    (6.2, 0.2, 1.6),
    (0.3, 0.2, 0),
    (0.3, 0.2, 0),
    (0.8, 0.2, 0),
    (0.8, 0.2, 0),
    (1.8, 0.2, 0),
    (1.8, 0.2, 0),
    (3, 0.2, 0),
    (3, 0.2, 0),
    (7.8, 0.2, 0),
    (7.8, 0.2, 0),
    (0.5, 0, 0),
    (1, 0, 0),
    (2, 0, 0),
    (3.2, 0, 0),
    (8, 0, 0),
    (0.1, 0, 0.4),
    (0.5, 0, 0.5),
    (1.4, 0, 0.6),
    (2.5, 0, 0.7),
    (7.3, 0, 0.7),
    (0, 0.1, 0.3),
    (0.1, 0.1, 0.7),
    (1, 0.1, 0.8),
    (2.1, 0.1, 1),
    (6.7, 0.1, 1.2),
    (0.1, 0.1, 0.3),
    (0.3, 0.1, 0.6),
    (0.7, 0.1, 1.2),
    (1.7, 0.1, 1.4),
    (6.1, 0.1, 1.8),
    (0.3, 0, 0.1),
    (0.7, 0, 0.3),
    (1.2, 0, 0.7),
    (2.4, 0, 0.8),
    (7, 0, 0.9),
    (0.5, 0, 0),
    (1, 0, 0),
    (1.5, 0, 0.5),
    (2.6, 0, 0.5),
    (7.4, 0, 0.6),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.4),
    (0, 0.1, 0.8),
    (0, 0.1, 0.8),
    (0, 0.1, 0.8),
    (0.4, 0.1, 1.4),
    (0.4, 0.1, 1.4),
    (0.4, 0.1, 1.4),
    (1.2, 0.1, 1.8),
    (1.2, 0.1, 1.8),
    (1.2, 0.1, 1.8),
    (5.3, 0.1, 2.6),
    (5.3, 0.1, 2.6),
    (5.3, 0.1, 2.6),
    (0, 0.1, 0.4),
    (0, 0.1, 0.8),
    (0.4, 0.1, 1.4),
    (1.2, 0.1, 1.9),
    (5.3, 0.1, 2.6),
    (0, 0.1, 0.4),
    (0.1, 0.1, 0.8),
    (0.6, 0.1, 1.3),
    (1.4, 0.1, 1.7),
    (5.6, 0.1, 2.2),
    (0, 0.2, 0.3),
    (0.3, 0.2, 0.5),
    (1, 0.2, 0.8),
    (2.1, 0.2, 0.9),
    (6.7, 0.2, 1.1),
    (0.3, 0.2, 0),
    (0.3, 0.2, 0),
    (0.8, 0.2, 0),
    (0.8, 0.2, 0),
    (1.8, 0.2, 0),
    (1.8, 0.2, 0),
    (3, 0.2, 0),
    (3, 0.2, 0),
    (7.8, 0.2, 0),
    (7.8, 0.2, 0)
]


def simulate(precip, tile_string):
    soil_type, land_use = tile_string.split(':')
    ki = lookup_ki(land_use)
    return simulate_cell_day((precip, 0.207 * ki), tile_string, 1)


def average(l):
    return reduce(lambda x, y: x + y, l) / len(l)


class TestModel(unittest.TestCase):
    """
    Model test set.
    """
    def test_nrcs(self):
        """
        Test the implementation of the runoff equation.
        """
        # This pair has CN=55
        runoffs = [round(runoff_nrcs(precip, 0.0, 'b', 'deciduous_forest'), 2)
                   for precip in PS]
        # Low curve number and low P cause too-high runoff
        self.assertEqual(runoffs[4:], CN55[4:])

        # This pair has CN=70
        runoffs = [round(runoff_nrcs(precip, 0.0, 'c', 'deciduous_forest'), 2)
                   for precip in PS]
        self.assertEqual(runoffs[1:], CN70[1:])

        # This pair has CN=80
        runoffs = [round(runoff_nrcs(precip, 0.0, 'd', 'pasture'), 2)
                   for precip in PS]
        self.assertEqual(runoffs, CN80)

        # This pair has CN=90
        runoffs = [round(runoff_nrcs(precip, 0.0, 'c', 'hi_residential'), 2)
                   for precip in PS]
        self.assertEqual(runoffs, CN90)

    def test_simulate_day_1(self):
        """
        Test the tile simulation using sample input/output.
        """
        # The number 0.04 is not very meaningful, this test just
        # attempts to give some idea about the mean error of the three
        # quantities, relative to precipitation, compared to the
        # sample output that was emailed to us.
        def similar(incoming, expected):
            precip, tile_string = incoming
            results = simulate(precip, tile_string)
            results = (results['runoff-vol'],
                       results['et-vol'],
                       results['inf-vol'])
            lam = lambda x, y: abs(x - y) / precip
            me = average(map(lam, results, expected))
            # Precipitation levels <= 2 inches are known to be
            # problematic.  It is unclear why the 'rock' type is
            # giving trouble on soil types C and D.
            if precip > 2 and tile_string != 'c:rock' \
               and tile_string != 'd:rock':
                self.assertTrue(me < 0.04, tile_string + ' ' + str(me))
        map(similar, INPUT, OUTPUT)

    def test_simulate_day_2(self):
        """
        Another test of the tile simulation using sample input/output.
        """
        # Test the RMSE of the runoff levels produced by the tile
        # simulation against values sample input/output.  The number
        # 0.13 is not very meaningful, this test just attempts to put
        # a bound on the deviation between the current output and the
        # sample output that was mailed to us.
        results = [simulate(precip, tile_string)['runoff-vol'] / precip
                   for precip, tile_string in INPUT
                   if precip > 2 and tile_string != 'c:rock' and
                   tile_string != 'd:rock']
        expected = [OUTPUT[i][0] / INPUT[i][0]
                    for i in range(len(INPUT))
                    if INPUT[i][0] > 2 and INPUT[i][1] != 'c:rock' and
                    INPUT[i][1] != 'd:rock']
        lam = lambda x, y: pow((x - y), 2)
        rmse = sqrt(average(map(lam, results, expected)))
        self.assertTrue(rmse < 0.13)

    def test_simulate_day_3(self):
        """
        Daily simulation.
        """
        result1 = simulate_cell_day((42, 93), 'a:rock', 1)
        result2 = simulate_cell_day((42, 93), 'a:rock', 2)
        self.assertEqual(result1['runoff-vol'] * 2, result2['runoff-vol'])

    def test_simulate_year(self):
        """
        Yearly simulation.
        """
        result1 = simulate_cell_year('a:hi_residential', 42, False)
        result2 = simulate_cell_year('a:mixed_forest', 42, False)
        self.assertNotEqual(result1, result2)
        self.assertGreater(result1['runoff-vol'], result2['runoff-vol'])

    def test_simulate_year_precolumbian(self):
        """
        Yearly simulation in Pre-Columbian times.
        """
        result1 = simulate_cell_year('a:hi_residential', 42, True)
        result2 = simulate_cell_year('a:mixed_forest', 42, True)
        self.assertEqual(result1, result2)

    def test_create_unmodified_census(self):
        """
        Test create_unmodified_census.
        """
        census = {
            "cell_count": 2,
            "distribution": {
                "a:rock": {"cell_count": 1},
                "a:water": {"cell_count": 1}
            },
            "modifications": [
                {
                    "bmp": "cluster_housing",
                    "cell_count": 1,
                    "distribution": {
                        "a:rock": {"cell_count": 1}
                    }
                }
            ]
        }

        result = create_unmodified_census(census)
        census.pop("modifications", None)
        self.assertEqual(census, result)

    def test_create_modified_census_1(self):
        """
        create_modified_census with a census tree without modifications.
        """
        census = {
            "cell_count": 5,
            "distribution": {
                "a:rock": {"cell_count": 3},
                "a:water": {"cell_count": 2}
            }
        }

        result = create_modified_census(census)
        census.pop("modifications", None)
        self.assertEqual(census, result)

    def test_create_modified_census_2(self):
        """
        create_modified_census with a census tree with trivial
        modifications.
        """
        census = {
            "cell_count": 3,
            "distribution": {
                "a:rock": {"cell_count": 2},
                "a:water": {"cell_count": 1}
            },
            "modifications": []
        }

        result = create_modified_census(census)
        census.pop("modifications", None)
        self.assertEqual(census, result)

    def test_create_modified_census_3(self):
        """
        create_modified_census with a census tree non-trivial
        modifications.
        """
        census1 = {
            "cell_count": 144,
            "distribution": {
                "a:rock": {"cell_count": 55},
                "a:water": {"cell_count": 89}
            },
            "modifications": [
                {
                    "bmp": "cluster_housing",
                    "cell_count": 34,
                    "distribution": {
                        "a:rock": {"cell_count": 34}
                    }
                }
            ]
        }

        census2 = {
            "cell_count": 144,
            "distribution": {
                "a:rock": {
                    "cell_count": 55,
                    "distribution": {
                        "a:cluster_housing": {"cell_count": 34},
                        "a:rock": {"cell_count": 21}
                    }
                },
                "a:water": {"cell_count": 89}
            }
        }

        result = create_modified_census(census1)
        self.assertEqual(census2, result)

    def test_simulate_water_quality_1(self):
        """
        Test the water quality simulation.
        """
        census = {
            "cell_count": 5,
            "distribution": {
                "a:rock": {"cell_count": 3},
                "a:water": {"cell_count": 2}
            }
        }

        def fn(cell, cell_count):
            return simulate_cell_year(cell, cell_count, False)

        simulate_water_quality(census, 93, fn)
        left = census['distribution']['a:rock']
        right = census['distribution']['a:water']
        for key in set(census.keys()) - set(['distribution']):
            self.assertEqual(left[key] + right[key], census[key])

    def test_simulate_water_quality_2(self):
        """
        Test the water quality simulation in the presence of modifications.
        """
        census = {
            "cell_count": 2,
            "distribution": {
                "a:rock": {"cell_count": 1},
                "a:water": {"cell_count": 1}
            },
            "modifications": [
                {
                    "reclassification": "d:hi_residential",
                    "cell_count": 1,
                    "distribution": {
                        "a:rock": {"cell_count": 1}
                    }
                }
            ]
        }

        census1 = create_modified_census(census)
        census2 = {
            "cell_count": 2,
            "distribution": {
                "d:hi_residential": {"cell_count": 1},
                "a:water": {"cell_count": 1}
            }
        }

        def fn(cell, cell_count):
            return simulate_cell_year(cell, cell_count, False)

        simulate_water_quality(census1, 93, fn)
        simulate_water_quality(census2, 93, fn)
        for key in set(census1.keys()) - set(['distribution']):
            self.assertEqual(census1[key], census2[key])

    def test_simulate_water_quality_precolumbian(self):
        """
        Test the water quality simulation in Pre-Columbian times.
        """
        census1 = {
            "cell_count": 3,
            "distribution": {
                "a:rock": {"cell_count": 1},
                "b:herbaceous_wetland": {"cell_count": 1},
                "a:water": {"cell_count": 1}
            }
        }

        census2 = {
            "cell_count": 3,
            "distribution": {
                "a:mixed_forest": {"cell_count": 1},
                "b:herbaceous_wetland": {"cell_count": 1},
                "a:water": {"cell_count": 1}
            }
        }

        def fn(cell, cell_count):
            return simulate_cell_year(cell, cell_count, True)

        simulate_water_quality(census1, 93, fn)
        simulate_water_quality(census2, 93, fn)
        for key in set(census1.keys()) - set(['distribution']):
            self.assertEqual(census1[key], census2[key])

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

"""
Operation tests.
"""

import unittest

from tr55.operations import dict_plus


class TestOperations(unittest.TestCase):
    """
    Dictionary operation test set.
    """
    def test_plus(self):
        """
        Test dictionary addition.
        """
        a = {'x': {'y': {'z': {'a': 1, 'b': 3, 'c': 13}, 'n': 144}}}
        b = {'x': {'y': {'z': {'a': 1, 'b': 5, 'c': 21}, 'm': 610}}}
        c = {'x': {'y': {'z': {'a': 2, 'b': 8, 'c': 34}, 'n': 144, 'm': 610}}}
        self.assertEqual(dict_plus(a, b), c)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

import sys


def tandem_walk(op, pred, left, right):
    """
    Walk two similarly-structured dictionaries in tandem, performing
    the given operation when both halves satisfy the given predicate.
    Otherwise, if both haves are dictionaries, recurse.
    """
    if pred(left) and pred(right):
        return op(left, right)
    elif isinstance(left, dict) and isinstance(right, dict):
        retval = left.copy()
        left_set = set(left.keys())
        right_set = set(right.keys())
        intersection = left_set & right_set
        difference = right_set - left_set
        for key in intersection:
            retval[key] = tandem_walk(op, pred, left[key], right[key])
        for key in difference:
            retval[key] = right[key]
        return retval


def isnumber(obj):
    """
    Is obj a number?
    """
    if sys.version_info.major == 3:
        types = (int, float)
    else:
        types = (int, long, float)
    return isinstance(obj, types)


def plus(x, y):
    """
    Sum of two numbers.
    """
    return x + y


def dict_plus(left, right):
    """
    Sum of two similarly-structured dictionaries.
    """
    return tandem_walk(plus, isnumber, left, right)
